#include "TPrgLine.h"
TPrgLine::TPrgLine(char* code):TObject()
{
  mCodeRaw = code;	
  mCode = "";	
  mLabel = "";	
	mComment = "";
	mTokens = 0;
	mId = 0;
	mIsExpanded = false;
	mParent = 0;
	mComp = 0;
	preStage();
	if(mId==kCOMMAND) expand();
}
TPrgLine::~TPrgLine()
{
  if(mTokens) delete mTokens;
	mExpand.Delete();
}
// this method removes spaces in the beginning and end of a String
void TPrgLine::cutSpaces(TString &L)
{
  while(L.Length()>0 && L.BeginsWith(" ")) L = L(1,L.Length()-1);
  while(L.Length()>0 && L.EndsWith(" "))   L = L(0,L.Length()-1);
}
void TPrgLine::cutTrash(TString &L)
{
  while(L.Length()>0 && (int)L[0]<32)              L = L(1,L.Length()-1);
  while(L.Length()>0 && (int)L[L.Length()-1]<32)   L = L(0,L.Length()-1);
}
// this method prestages the line by doing
// formatting
void TPrgLine::preStage()
{
  // replace Tabs by space
	mCodeRaw.ReplaceAll("\t"," ");	
	cutTrash(mCodeRaw);
	
	TString a = mCodeRaw;	
		
	if(a.Length()<=0) return;
		
	// look for comments in the line
	int i = a.Index(";");
  if(i>=0) 
	{ 
		mComment = a(i,a.Length()-i);
	  a = a(0,i);
	}
	
	mId = kCOMMENT;	
	if(a.Length()<=0) return;

	// To upper case. From now on do not make difference.
	a.ToUpper();
	
	// check any of the preCompiler options in the old
	// ADAC format. 
	// pre-compiling options are not necessary in the
	// ADAC+
	
  if(a.BeginsWith("PGM$") ||
	   a.BeginsWith("COM$") ||
	   a.BeginsWith("COMANDO") ||
		 a.BeginsWith("CSYS") ||
		 a.BeginsWith("CREM") ||
		 a.BeginsWith("CAUX") ||
		 a.BeginsWith("NAUX") ||
		 a.BeginsWith("NSHD")) { mId = kPRECOMP; mCode = a; return;}
 		
	// checks is there is a label. Labels should start in the
	// first character of the line. If the first character is
	// a space we assume there is no label
	// labels can be as long as the user wishes.
	// after the label is found (or not) we remove
	// extra space characters in the string
	if(a.BeginsWith(" "))
	{
	  // regular command, no label
		cutSpaces(a);
	}
	else
	{
	  i = a.Index(" ");
		mLabel = a(0,i);
		a = a(i,a.Length()-i);
		cutSpaces(a);
	}
	
	if(a.Length()<=0) return;
	
	// check if it is a symbol operation. Symbol definitions
	// and operations are identified by an equate '=' symbol
	// in the expression.
	i = a.Index("=");
	
	mCode = a;
	
	if(i>=0) { mId =  kSYMBOL; return;}
	
	// it is not a symbol so should be a regular command
	mId = kCOMMAND;			
    
  mCode.ReplaceAll("["," [ ");
  mCode.ReplaceAll("]"," ] ");
  mCode.ReplaceAll("."," ");
  mCode.ReplaceAll(","," ");
  mCode.ReplaceAll("("," ");
  mCode.ReplaceAll(")"," ");
  mCode.ReplaceAll("="," = ");
  mCode.ReplaceAll("+"," + ");
  mCode.ReplaceAll("-"," - ");
  mCode.ReplaceAll("*"," * ");
  mCode.ReplaceAll("/"," / ");
	
	TString tokens = " ";
	mTokens = mCode.Tokenize(tokens);
}
void TPrgLine::addExpanded(char* line)
{
  TPrgLine *P = new TPrgLine(line);
	mExpand.Add(P);
	P->setParent(this);
}
// this method checks if the command is 
// a macro command and expands it if necessary
// macro commands are: CNAF, CNAFX, IF, ROUT, CROUT, ROUTX, CROUTX, SPAT, CSPAT
void TPrgLine::expand()
{
  char* CMD = getToken(0);
	if(!CMD) return;
	
	char line[300];
	
	if(!strcmp(CMD,"CNAF"))
	{
	  mIsExpanded = true;
		sprintf(line,"%s SETC %s",mLabel.Data(), getToken(1));
		addExpanded(line);
		sprintf(line," NAF %s %s %s %s", getToken(2), getToken(3), getToken(4), getToken(5));
		addExpanded(line);		
	}
	if(!strcmp(CMD,"CNAFX"))
	{
	  mIsExpanded = true;
		sprintf(line,"%s SETC %s",mLabel.Data(), getToken(1));
		addExpanded(line);
		sprintf(line," NAFX %s %s %s %s", getToken(2), getToken(3), getToken(4), getToken(5));
		addExpanded(line);		
	}
	if(!strcmp(CMD,"IF"))
	{
	  mIsExpanded = true;
		char*   OP = getToken(2);
		TString PO = "NOTFOUND";
		
		if(!strcmp(OP,"ANY")) PO = "NONE";
		if(!strcmp(OP,"NONE")) PO = "ANY";
		if(!strcmp(OP,"GE")) PO = "LT";
		if(!strcmp(OP,"LE")) PO = "GT";
		if(!strcmp(OP,"GT")) PO = "LE";
		if(!strcmp(OP,"LT")) PO = "GE";
		if(!strcmp(OP,"NE")) PO = "EQ";
		if(!strcmp(OP,"EQ")) PO = "NE";
		
		sprintf(line,"%s SKIP %s %s",mLabel.Data(), getToken(1), PO.Data());
		for(int i = 3; i<getNTokens()-1; i++) sprintf(line,"%s %s",line,getToken(i));
		addExpanded(line);
		sprintf(line," BRU %s", getToken(getNTokens()-1));
		addExpanded(line);
	}
	if(!strcmp(CMD,"ROUT"))
	{
	  mIsExpanded = true;
		TString C = "CA";
		if(getToken(4)) C = getToken(4);
		sprintf(line,"%s NAF %s %s %s %s",mLabel.Data(), getToken(1), getToken(2), getToken(3), C.Data());
		addExpanded(line);		
		sprintf(line," OUT %s",C.Data());
		addExpanded(line);		
	}
	if(!strcmp(CMD,"CROUT"))
	{
	  mIsExpanded = true;
		TString C = "CA";
		if(getToken(5)) C = getToken(5);
		sprintf(line,"%s SETC %s",mLabel.Data(), getToken(1));
		addExpanded(line);		
		sprintf(line," NAF %s %s %s %s", getToken(2), getToken(3), getToken(4), C.Data());
		addExpanded(line);		
		sprintf(line," OUT %s",C.Data());
		addExpanded(line);		
	}
	if(!strcmp(CMD,"ROUTX"))
	{
	  mIsExpanded = true;
		TString C = "CA";
		if(getToken(4)) C = getToken(4);
		sprintf(line,"%s NAFX %s %s %s %s",mLabel.Data(), getToken(1), getToken(2), getToken(3), C.Data());
		addExpanded(line);		
		sprintf(line," OUT %s",C.Data());
		addExpanded(line);		
	}
	if(!strcmp(CMD,"CROUTX"))
	{
	  mIsExpanded = true;
		TString C = "CA";
		if(getToken(5)) C = getToken(5);
		sprintf(line,"%s SETC %s",mLabel.Data(), getToken(1));
		addExpanded(line);		
		sprintf(line," NAFX %s %s %s %s", getToken(2), getToken(3), getToken(4), C.Data());
		addExpanded(line);		
		sprintf(line," OUT %s",C.Data());
		addExpanded(line);		
	}
	if(!strcmp(CMD,"SPAT"))
	{
	  mIsExpanded = true;
		TString C = "CA";
		if(getToken(4)) C = getToken(4);
		sprintf(line," NAF %s %s %s %s", getToken(1), getToken(2), getToken(3), C.Data());
		addExpanded(line);		
		sprintf(line," MOV %s PAT",C.Data());
		addExpanded(line);		
	}
	if(!strcmp(CMD,"CSPAT"))
	{
	  mIsExpanded = true;
		TString C = "CA";
		if(getToken(5)) C = getToken(5);
		sprintf(line,"%s SETC %s",mLabel.Data(), getToken(1));
		addExpanded(line);		
		sprintf(line," NAF %s %s %s %s", getToken(2), getToken(3), getToken(4), C.Data());
		addExpanded(line);		
		sprintf(line," MOV %s PAT",C.Data());
		addExpanded(line);		
	}
}
int TPrgLine::getNTokens()
{
  if (mTokens) return mTokens->GetEntries(); 
	else return 0;
}
char* TPrgLine::getToken(int i)
{
  if (getNTokens()>0) 
	{
	  if(i<getNTokens())
		{
		  return (char*)((TObjString*)mTokens->At(i))->GetString().Data();
		}
	}	
	return NULL;
}
void TPrgLine::print(bool expand)
{
	char line[400];
	if(expand)
	{
	  sprintf(line,"%05d : %s",mId, mCodeRaw.Data());
	  cout <<"----------------------------------------"<<endl;
	  cout <<line<<endl;
	  cout <<"  Number of tokens = "<<getNTokens()<<endl;
	  cout <<"  Label = "<<mLabel.Data()<<endl;
	  cout <<"  Code = "<<mCode.Data()<<endl;
	  cout <<"  Comment = "<<mComment.Data()<<endl;
	  cout <<"  Id = "<<mId<<endl;
		cout <<"  Compilation code = "<<mComp<<endl;;
	  cout <<"  Tokens = ";
	  for(int i = 0; i<getNTokens(); i++)
	  {
	    cout << getToken(i)<<" | ";
	  }
	  cout <<endl;
	  if(isExpanded())
	  {
	    cout <<"***** Command is expanded **************************"<<endl;
	  	int n = getNExpanded();
	  	for(int i = 0;i<n;i++) getExpanded(i)->print();
	    cout <<"****************************************************"<<endl;
	  }
	}
	else
	{
		if(getId()==kCOMMENT)
			sprintf(line,"%05d | %02d | %08d | %s",getLineNumber(),getId(),0,getComment());
		else
			sprintf(line,"%05d | %02d | %08x | %8s  %s %s",getLineNumber(),getId(),getCompiledCode(),getLabel(),getCode(),getComment());
	  cout <<line<<endl;
	}
	
}





